/**************************************************************************** 
*  Copyright 2008 C&H Technologies, Inc.  All Rights Reserved.					    * 
*****************************************************************************/
/****************************************************************************
*
*	m220_ansi.c  
*
*	ANSI-C source module for the M220
*
* This file contains high level functions to control a M220 
* This file is written entirely in ANSI-C so that is may be compiled in
* any environment with a supporting ANIS-C compiler It uses the APIS 
* library to perform instrument communications.  
*
*	This file is organized such that modifications are easy.  All functions 
*	that might need to be modified are located at the end of this file.  
*	Comments are provided that explain tasks of the function and how it might 
*	be modified.  All other function should not need to be modified.
*
*	C&H Part Number: 11029557
*
*	Revision History
*	-	Fixed m220_RouteCloseQuery() so that is returns a 0 or 1.  Previously it
*		was returning the corresponding bit of the corresponding Set/Reset 
*		register.  It returned 0,1,2,4, or 8.  (David Clark 06/03/08)
*	-	Modified m220_RouteClose() so that it only does a single register write.
*		Previously it was performing an unecessary read-modify-write.  
*		(David Clark 06/03/08)
*	-	Fixed m220_RouteOpenQuery() so that is returns a 0 or 1.  Previously it
*		was returning the corresponding bit of the corresponding Set/Reset 
*		register.  It returned 0,1,2,4, or 8.  (David Clark 06/03/08)
*	-	Modified m220_RouteOpen() so that it only does a single register write.
*		Previously it was performing an unecessary read-modify-write.  
*		(David Clark 06/03/08)
*	-	In m220_RouteCloseList() and m220_RouteCloseList() we set overlap flag 
*		to 0 temporarirly.  Not doing so caused some Relay Close/Open to be lost
*		due to FIFO full.  
*		(David Clark 06/03/08)
*	-	Clear the list string near the beginning of the functions
*		m220_RouteCloseListQuery() and m220_RouteOpenListQuery().  These funcitons 
*		uses strcat() and if there was 'junk' in the string before calling these 
*		functions, it will remain and the list will be concatenated to it.
*	-	Fixed m220_ChannelStatus() so that is returns a 0 or 1.  Previously it
*		was returning the corresponding bit of the corresponding Set/Reset 
*		register.  It returned 0,1,2,4, or 8.  (David Clark 06/03/08)
*	-	Changed m220_SetAllQuery() so that is read each Set/Reset Register and 
*		creates the bit-field from that rather can going bit by bit.  This wasn't
*		a bug be this change makes the code simpler and improves performance.
*		(David Clark 06/03/08)
*	-	Changed m220_SetAll() so that it creates the bit values for each Set/Reset
*		register then performs the writes.  Previously it simply called 
*		m220_RouteClose() and m220_RouteOpen().  This caused problem if overlap was
*		1 because performing so many back to back write would cause some Open/Close
*		commands to be lost due to FIFO Full.
*	-	Release Version 1.1 (David Clark 06/03/08)   
*
*****************************************************************************/
#include <ansi_c.h>
#include "m220_ansi.h"

/**************************************************************************** 
*  Non Exported Defines
*****************************************************************************/
/*  Driver Revision  */
#define		M220_DRIVER_REV			"1.1"

/**************************************************************************** 
*  Error Checking Macros
*****************************************************************************/
#ifndef checkParamRange
#define  checkParamRange(value, min, max, param)		if ((value > max) || (value < min)) { \
																											error = M220_ERROR_PARAMETER1 + param-1;\
																											goto Error; \
																										} else 
#endif																				
																				

#ifndef checkParamPointer
#define  checkParamPointer(pointer, param)		if (pointer == NULL) { \
																							error = M220_ERROR_PARAMETER1 + param-1;\
																							goto Error; \
																							} else 
#endif																				

																				
#ifndef checkErr
#define checkErr(fCall)      if (error = (fCall), (error = (error < 0) ? error : M220_SUCCESS)) \
                                 {goto Error;}  else
#endif

#ifndef apisCheckErr
#define apisCheckErr(fCall)      if (error = (fCall), (error = (error != 0) ? error : M220_SUCCESS)) \
                                 {	handle->apis_error = error;\
																		error = M220_ERROR_APIS; \
																		goto Error;}  else
#endif

#ifndef apisCheckErrInit
#define apisCheckErrInit(fCall)      if (error = (fCall), (error = (error != 0) ? error : M220_SUCCESS)) \
                                 {	(*handle)->apis_error = error;\
																		error = M220_ERROR_APIS; \
																		goto Error;}  else
#endif

/**************************************************************************** 
*  Constants
*****************************************************************************/
static const char *errorMessageTable[M220_SPECIFIC_ERROR_END - M220_SPECIFIC_ERROR_BASE] = 
						{	
							"Wait timeout expired while waiting for the operation to complete.",
							"An invalid parameter was passed to the function in parameter 1",
							"An invalid parameter was passed to the function in parameter 2",
							"An invalid parameter was passed to the function in parameter 3",
							"An invalid parameter was passed to the function in parameter 4",
							"The module returned an invalid ID.",
							"An error occurred in the APIS Platform Support Module.  Use the function ma204_GetApisError() to retrieve the APIS error code.",
							"The operation could not be completed because the system is out of memory."
							"The attribute specified is not valid for this driver."
						};


static const char *warningMessageTable[M220_SPECIFIC_WARNING_END - M220_SPECIFIC_WARNING_BASE] = 
						{	""
						};



/**************************************************************************** 
*  Non Exported Functions
*****************************************************************************/
static Int32 m220_IDWriteByte(M220_HANDLE handle, UInt8 value);
static Int32 m220_IDReadByte(M220_HANDLE handle, UInt8 *value);
static Int32 m220_IDWriteBit(M220_HANDLE handle, Bool value);
static Int32 m220_IDReadBit(M220_HANDLE handle, Bool *value);


/*****************************************************************************
*	m220_RevisionQuery
*
*	Returns the version number of this driver.  
******************************************************************************/
Int32 m220_RevisionQuery (Int8 driver_rev[])
{	
	Int32 error = M220_SUCCESS;

/*  Check parameters  */
	checkParamPointer(driver_rev, 2);
	
	sprintf(driver_rev, "%s", M220_DRIVER_REV);
Error:
	return error;
}

/*****************************************************************************
* m220_ErrorMessage()
*
* Translates error codes to user readable error message
******************************************************************************/
Int32 m220_ErrorMessage(Int32 statusCode, char errorMessage[])
{	
	Int32 error = M220_SUCCESS;
	
/*  Check parameters  */
	checkParamPointer(errorMessage, 3);

/*  Check driver defined error codes  */
	if((statusCode >= M220_SPECIFIC_ERROR_BASE) && (statusCode < M220_SPECIFIC_ERROR_END))
	{	strcpy(errorMessage, errorMessageTable[statusCode - M220_SPECIFIC_ERROR_BASE]);
	}
	else if((statusCode >= M220_SPECIFIC_WARNING_BASE) && (statusCode < M220_SPECIFIC_WARNING_END))
	{	strcpy(errorMessage, warningMessageTable[statusCode - M220_SPECIFIC_WARNING_BASE]);
	}
	else if (statusCode == 0)
		sprintf(errorMessage, "No Error Occured");
	else
		sprintf(errorMessage, "Unknown Error Code [0x%x]", statusCode);
		
Error:
	return error;
}


/*****************************************************************************
*	m220_GetIdentWord
*
*	Reads the specified offset from the module's ID space and returns
*	the result.
******************************************************************************/
Int32 m220_GetIdentWord (M220_HANDLE handle, UInt16 offset, UInt16 *value)
{	
	Int32 error = M220_SUCCESS;
	UInt8 data;

/*  Check parameters  */
	checkParamPointer(handle, 1);
	checkParamPointer(value, 3);

	*value = 0;
	
/*  Send the read command to the serial EEPROM  */	
	checkErr(m220_IDWriteByte(handle, 0x80 | offset));
	
/*  Read upper byte  */	
	checkErr(m220_IDReadByte(handle, &data));
	*value = data << 8;
	
/*  Read lower byte  */	
	checkErr(m220_IDReadByte(handle, &data));
	*value |= data;
	
Error:	
	m220_WriteRegister(handle, M220_ID_ACCESS_REGISTER, 0);
	return error;
}

/*****************************************************************************
* m220_SetTimerMode
*
* Sets the timer
* This is the amount of time that power will be applied from the module
* to the relay coils.
*
* Timer Mode (proper operation is guaranteed only for the time = 8ms)
*				0x00	8ms
*				0x01	2ms
*				0x10	4ms
*				0x11	64ms
******************************************************************************/
Int32 m220_SetTimerMode(M220_HANDLE handle, UInt16 mode)
{ 
	Int32 error = M220_SUCCESS;
	UInt16 control_reg;
	UInt16 data;       

	checkParamPointer(handle, 1);
	
	checkErr(m220_ReadRegister (handle, M220_CONTROL_REGISTER, &control_reg));

	control_reg &= ~M220_CONTROL_TIMER_BITS;												 
	data = (mode << 4) & M220_CONTROL_TIMER_BITS;              		   
	control_reg |= data;																						
	
	checkErr(m220_WriteRegister (handle, M220_CONTROL_REGISTER, control_reg));

Error:
	return error;
}


/*****************************************************************************
*	m220_EnableRelayPower
*	
* Enables power to relays
******************************************************************************/
Int32 m220_EnableRelayPower(M220_HANDLE handle, Bool enable)
{ 
	Int32 error = M220_SUCCESS;
	UInt16 control_reg;
	UInt16 data;
	
	checkParamPointer(handle, 1);

	checkErr(m220_ReadRegister (handle, M220_CONTROL_REGISTER, &control_reg));

	control_reg &= ~M220_CONTROL_DRIVER_POWER_ENABLE_BIT;						 
	data = (enable << 3) & M220_CONTROL_DRIVER_POWER_ENABLE_BIT;     
	control_reg |= data;																							 
	
	checkErr(m220_WriteRegister (handle, M220_CONTROL_REGISTER, control_reg));

Error:
	return error;
}


/*****************************************************************************
*	m220_EnableSelfTest
*	
* Self-Test Enable:  Disables power to all row & column drivers and allows testing 
* the operation of the FIFO without causing relay movement, 
******************************************************************************/
Int32 m220_EnableSelfTest(M220_HANDLE handle, Bool enable)
{ 
	Int32 error = M220_SUCCESS;
	UInt16 control_reg;
	UInt16 data;

	checkParamPointer(handle, 1);
	
	checkErr(m220_ReadRegister (handle, M220_CONTROL_REGISTER, &control_reg));            

	control_reg &= ~M220_CONTROL_SELFTEST_ENABLE_BIT;						              
	data = (enable << 2) & M220_CONTROL_SELFTEST_ENABLE_BIT;              		  
	control_reg |= data;																							             
	
	checkErr(m220_WriteRegister (handle, M220_CONTROL_REGISTER, control_reg));		

Error:
	return error;
}

/*****************************************************************************
*	m220_EnableInterrupts 
*
******************************************************************************/
Int32 m220_EnableInterrupts (M220_HANDLE handle, Bool enable)
{	
	Int32 error = M220_SUCCESS;
	UInt16 control_reg;
	UInt16 data;

	checkParamPointer(handle, 1);
	checkParamRange(enable, 0, 1, 2);
	
	checkErr(m220_ReadRegister (handle, M220_CONTROL_REGISTER, &control_reg));

	control_reg &= ~M220_CONTROL_INTERRUPT_ENABLE_BIT;
	data = (enable << 1) & M220_CONTROL_INTERRUPT_ENABLE_BIT;
	control_reg |= data;

	checkErr(m220_WriteRegister (handle, M220_CONTROL_REGISTER, control_reg));
	
Error:
	return error;
}


/*****************************************************************************
*	m220_InitStatus
*	
*	Returns the status of the device initialization
******************************************************************************/
Int32 m220_InitStatus(M220_HANDLE handle, Bool *status)
{	
	Int32 error = M220_SUCCESS;
	UInt16 stat_reg_value;

	checkParamPointer(handle, 1);
	checkParamPointer(status, 2);
	
	checkErr(m220_ReadRegister (handle, M220_STATUS_REGISTER, &stat_reg_value));
	*status = (Bool)(stat_reg_value & M220_INIT_STATUS_BIT);

Error:
	return error;
}

/*****************************************************************************
*	m220_FIFOEmptyStatus
*	
*	Returns the status of the FIFO buffer
******************************************************************************/
Int32 m220_FIFOEmptyStatus(M220_HANDLE handle, Bool *status)
{	
	Int32 error = M220_SUCCESS;
	UInt16 stat_reg_value; 

	checkParamPointer(handle, 1);
	checkParamPointer(status, 2);

	checkErr(m220_ReadRegister (handle, M220_STATUS_REGISTER, &stat_reg_value));
	*status = (Bool)(stat_reg_value & M220_FIFO_EMPTY_STATUS_BIT);

Error:
	return error;
}

/*****************************************************************************
*	m220_FIFOFullStatus
*	
*	Returns the status of the FIFO buffer
******************************************************************************/
Int32 m220_FIFOFullStatus(M220_HANDLE handle, Bool *status)
{	
	Int32 error = M220_SUCCESS;
	UInt16 stat_reg_value = 0;

	checkParamPointer(handle, 1);
	checkParamPointer(status, 2);

	checkErr(m220_ReadRegister (handle, M220_STATUS_REGISTER, &stat_reg_value));
	*status = (Bool)((stat_reg_value) & (M220_FIFO_FULL_STATUS_BIT));

Error:
	return error;
}

/*****************************************************************************
*	m220_InterruptAssertedStatus
*	
* Interrupt is asserted after the timer (relay settling time) has elapsed
******************************************************************************/
Int32 m220_InterruptAssertedStatus(M220_HANDLE handle, Bool *status)
{	
	Int32 error = M220_SUCCESS;
	UInt16 stat_reg_value;

	checkParamPointer(handle, 1);
	checkParamPointer(status, 2);

	checkErr(m220_ReadRegister (handle, M220_STATUS_REGISTER, &stat_reg_value));
	*status = (Bool)(stat_reg_value & M220_INTERRUPT_STATUS_BIT);   

Error:
	return error;
}


/*****************************************************************************
* m220_GetAttribute()
*
* Gets attribute value of
* timeout, overlap
******************************************************************************/
Int32 m220_GetAttribute(M220_HANDLE handle, Int16 attributeName, Int32 *attributeValue)
{ 
	Int32 error = M220_SUCCESS;


	checkParamPointer(handle, 1);
	checkParamPointer(attributeValue, 3);

	switch(attributeName)
	{
		case M220_ATTR_OVERLAP_MODE:
			*attributeValue = handle->overlap;
			break;
		case M220_ATTR_TIMEOUT:
			*attributeValue = handle->timeout_ms;
			break;
		case M220_ATTR_FIFO_FULL_WAIT:
			*attributeValue = handle->fifo_full_wait;
			break;
		default:
			error = M220_ERROR_INVALID_ATTRIBUTE;
			break;
	}
		
Error:
	return error;
}


/*****************************************************************************
* m220_SetAttribute()
*
* Sets attributes of
* timeout, overlap
******************************************************************************/
Int32 m220_SetAttribute(M220_HANDLE handle, Int16 attributeName, Int32 attributeValue)
{ 
	Int32 error = M220_SUCCESS;

	checkParamPointer(handle, 1);

	switch(attributeName)
	{
		case M220_ATTR_OVERLAP_MODE:
			checkParamRange(attributeValue, 0, 1, 3);
			handle->overlap = attributeValue;
			break;
		case M220_ATTR_TIMEOUT:
			handle->timeout_ms = attributeValue;
			break;
		case M220_ATTR_FIFO_FULL_WAIT:
			handle->fifo_full_wait = attributeValue;
			break;
		default:
			error = M220_ERROR_INVALID_ATTRIBUTE;
			break;
	}

Error:
	return error;
}

/*****************************************************************************
* m220_GetMUXMode()
*
*   mode = 1 for dual 8x1
*   mode = 0 for single 16x1
******************************************************************************/
Int32 m220_GetMUXMode(M220_HANDLE handle, Bool *mode)
{
	Int32 error = M220_SUCCESS;
	UInt16 reg;
	
	checkParamPointer(handle, 1);
	checkParamPointer(mode, 2);
	
	checkErr(m220_ReadRegister(handle, M220_STATUS_REGISTER, &reg));
	if(reg & M220_MPS_BIT)
		*mode = 1;
	else
		*mode = 0;
	
Error:
	return error;
	
}

/*****************************************************************************
* m220_OpcQuery()
*
* Returns 1 when operation complete (FIFO is Empty)
*
*  Note: Use when overlap = 0 in between
*        open / close relay operations to
*        insure settling time of relays has
*        elapsed. (~13 milliseconds)
******************************************************************************/
Int32 m220_OpcQuery(M220_HANDLE handle, Int16 *state )
{ 
	Int32 error = M220_SUCCESS;
	UInt16 busy;
	
	checkParamPointer(handle, 1);
	checkParamPointer(state, 2);
	
	checkErr(m220_ReadRegister(handle, M220_STATUS_REGISTER, &busy));
		
	if (busy & M220_FIFO_EMPTY_STATUS_BIT)
					*state = 1;
	else
					*state = 0;

Error:
	return error;
}

/*****************************************************************************
*	m220_IDWriteBit
*
*	Write a single bit to the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m220_IDWriteBit(M220_HANDLE handle, Bool value)
{	
	Int32 error = M220_SUCCESS;

/*  Static functions do not need parameter checks because we are
		calling them internally therefore we are garunteed valid parameters 
*/		
		
/*  data with clock low  */
	checkErr(m220_WriteRegister(handle, M220_ID_ACCESS_REGISTER, (0x4 | (value & 0x1))));

/*  data with clock high  */	 
	checkErr(m220_WriteRegister(handle, M220_ID_ACCESS_REGISTER, (0x6 | (value & 0x1))));
Error:	
	return error;
}


/*****************************************************************************
*	m220_IDWriteByte
*
*	Write a command byte to the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m220_IDWriteByte(M220_HANDLE handle, UInt8 value)
{	
	Int32 error = M220_SUCCESS;
	Int32 i;
	
/*  Static functions do not need parameter checks because we are
		calling them internally therefore we are garunteed valid parameters 
*/		
	
/*  Initialize  */	
	checkErr(m220_WriteRegister(handle, M220_ID_ACCESS_REGISTER, 0));
	checkErr(m220_WriteRegister(handle, M220_ID_ACCESS_REGISTER, 0x4));
	
/*  Write start bit  */	
	checkErr(m220_IDWriteBit(handle, 1));
	for(i=0; i<=7; i++)
	{
	/*  Write data bits  */		
		checkErr(m220_IDWriteBit(handle, ((value &0x80)>>7)));
		value = value << 1;
	}
Error:	
	return error;
}



/*****************************************************************************
*	ma220_IDReadBit
*
*	Reads a single bit from the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m220_IDReadBit(M220_HANDLE handle, Bool *value)
{	
	Int32 error = M220_SUCCESS;
	UInt16 data;

/*  Static functions do not need parameter checks because we are
		calling them internally therefore we are garunteed valid parameters 
*/		
	
/*  clock low  */ 	
	checkErr(m220_WriteRegister(handle, M220_ID_ACCESS_REGISTER, 0x4));
	
/*  clock high  */	
	checkErr(m220_WriteRegister(handle, M220_ID_ACCESS_REGISTER, 0x6));
	
/*  Read data  */	
	checkErr(m220_ReadRegister(handle, M220_ID_ACCESS_REGISTER, &data));
	*value = (Bool)data;

Error:	
	return error;
}


/*****************************************************************************
*	m220_IDReadByte
*
*	Reads a byte from the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m220_IDReadByte(M220_HANDLE handle, UInt8 *value)
{	
	Int32 error = M220_SUCCESS;
	Int8 i;
	Bool data;
	
/*  Static functions do not need parameter checks because we are
		calling them internally therefore we are garunteed valid parameters 
*/		
	
	*value = 0;
	for(i=7; i>=0; i--)
	{ 
	/*  read data 1 bit at a time  */	
		checkErr(m220_IDReadBit(handle, &data));
		*value |= ((data & 0x1) << i);
	}
	
Error:	
	return error;
}


/*****************************************************************************
*	m220_RouteCloseQuery
*
*****************************************************************************/
Int32 m220_RouteCloseQuery(M220_HANDLE handle, Int16 channel, Bool *closed)
{
	Int32 error = M220_SUCCESS;
	UInt16 offset, bit_mask, reg, result;
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, 15, 2);
	checkParamPointer(closed, 3);
	
	result = 0;
	offset = M220_ROW0_SET_REGISTER + (channel/4) * 4;
	
	bit_mask = 0x01 << (channel%4);
	          									
	checkErr(m220_ReadRegister (handle, offset, &reg));
	
	if(reg & bit_mask)
		*closed = M220_TRUE;
	else
		*closed = M220_FALSE;
	
Error:
	return error;
}


/*****************************************************************************
* m220_RouteClose
*
* Closes relay indicated by channel parameter
*****************************************************************************/
Int32 m220_RouteClose(M220_HANDLE handle, UInt16 channel)
{ 
	Int32 error = M220_SUCCESS;
	UInt16 offset, bit_mask;
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, 15, 2);
	
	//  Determine register offset and bit mask for channel
	//////////////////////////////////////////////////////////
	bit_mask = (0x01 << channel%4);
	offset = M220_ROW0_SET_REGISTER + ((channel/4) * 4);  // set registers are 4 bytes apart
	
	checkErr(m220_WriteRegister (handle, offset, bit_mask));
	
Error:
	return error;

}


/*****************************************************************************
*	m220_RouteOpenQuery
*
*****************************************************************************/
Int32 m220_RouteOpenQuery(M220_HANDLE handle, Int16 channel, Bool *opened)
{
	Int32 error = M220_SUCCESS;
	UInt16 offset, bit_mask, reg, result;
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, 15, 2);
	checkParamPointer(opened, 3);
	
	offset = M220_ROW0_RESET_REGISTER + ((channel/4) * 4);          						
	
	bit_mask = (0x01 << (channel%4));
	          									
	checkErr(m220_ReadRegister (handle, offset, &reg));
	
	if((~reg) & bit_mask)
		*opened = M220_TRUE;
	else
		*opened = M220_FALSE;
	
Error:
	return error;
}


/******************************************************************************
* m220_RouteOpen
*
* Opens relay indicated by channel
******************************************************************************/
Int32 m220_RouteOpen(M220_HANDLE handle, UInt16 channel)
{ 
	Int32 error = M220_SUCCESS;
	UInt16 bit_mask;
	UInt16 offset;
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, 15, 2);

	//  Determine register offset and bit mask for channel
	//////////////////////////////////////////////////////////
	bit_mask = (0x01 << channel%4);
	offset = M220_ROW0_RESET_REGISTER + ((channel/4) * 4);     // registers are 4 bytes apart

	checkErr(m220_WriteRegister (handle, offset, ~bit_mask));
Error:
	return error;

}

/******************************************************************************
* m220_RouteCloseList
*
* Closes a list of channels passed as a comma separated string (char *list)
******************************************************************************/
Int32 m220_RouteCloseList(M220_HANDLE handle, char *channel_list)
{ 
	Int32 error = M220_SUCCESS;
	Int32 chan, start_channel, end_channel;
	char *tokenptr;
	char *channel_list2;
	Bool saved_overlap;
	
	
	saved_overlap = handle->overlap;
	handle->overlap = 0;

	checkParamPointer(handle, 1);
	checkParamPointer(channel_list, 2);
	
	channel_list2 = malloc(256);

	strcpy(channel_list2, channel_list);
	
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			sscanf(tokenptr, "%d", &start_channel);
			end_channel = start_channel;
		}
		else
		{
			sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
		}
				
		for(chan = start_channel; chan <= end_channel; chan++) 
		{
			checkErr(m220_RouteClose(handle, chan));
		}
					
		tokenptr = strtok(NULL, ",");
			
	}
	
Error:
	handle->overlap = saved_overlap;
	return error;
}


/******************************************************************************
*	m220_RouteCloseListQuery
*
******************************************************************************/
Int32 m220_RouteCloseListQuery(M220_HANDLE handle, char *channel_list, char *route_close_list)
{
	Int32 error = M220_SUCCESS;
	UInt16 reg, offset;
	Int32 chan, start_channel, end_channel;
	Bool closed;
	char *tokenptr;
	char str[20];
	
	checkParamPointer(handle, 1);
	checkParamPointer(channel_list, 2);
	checkParamPointer(route_close_list, 3);
	
	route_close_list[0] = '\0';	
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			sscanf(tokenptr, "%d", &start_channel);
			end_channel = start_channel;
		}
		else
		{
			sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
		}
		
		for(chan = start_channel; chan <= end_channel; chan++) 
		{
			checkErr(m220_RouteCloseQuery(handle, chan, &closed));

			// if (closed), append 1-comma to list, otherwise append 0-comma
			if(closed)
			{
				sprintf(str, "%d,", 1);
				strcat(route_close_list, str);
			}
			else
			{
				sprintf(str, "%d,", 0);
				strcat(route_close_list, str);
			}
		}
		tokenptr = strtok(NULL, ",");
	}
Error:
	return error;
}


/**************************************************************************
* m220_RouteOpenList
*
* Closes a list of channels passed as a comma separated string (char *list)
*  - List may also have a colon to denote a range (eg - "2:8" for channels
*    2 through 8.
***************************************************************************/
Int32 m220_RouteOpenList(M220_HANDLE handle, char *channel_list)
{
	Int32 error = M220_SUCCESS;
	Int32 chan, start_channel, end_channel;
	char *tokenptr;
	char *channel_list2;
	Bool saved_overlap;
	
	saved_overlap = handle->overlap;
	handle->overlap = 0;

	checkParamPointer(handle, 1);
	checkParamPointer(channel_list, 2);
	
	channel_list2 = malloc(256);

	strcpy(channel_list2, channel_list);
	
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			sscanf(tokenptr, "%d", &start_channel);
			end_channel = start_channel;
		}
		else
		{
			sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
		}
				
		for(chan = start_channel; chan <= end_channel; chan++) 
		{
			checkErr(m220_RouteOpen(handle, chan));
		}
		tokenptr = strtok(NULL, ",");
	}
Error:
	handle->overlap = saved_overlap;
	return error;
}


/***************************************************************************** 
*	m220_RouteOpenListQuery 
******************************************************************************/
Int32 m220_RouteOpenListQuery(M220_HANDLE handle, char *channel_list, char *route_open_list)
{
	Int32 error = M220_SUCCESS;
	UInt16 reg, offset;
	Int32 chan, start_channel, end_channel;
	Bool opened;
	char *tokenptr;
	char str[20];
	
	checkParamPointer(handle, 1);
	checkParamPointer(channel_list, 2);
	checkParamPointer(route_open_list, 3);
	
	route_open_list[0] = '\0';
	tokenptr = strtok(channel_list, ",");
	while(tokenptr != NULL)
	{ 
		if(strchr(tokenptr, ':') == NULL)
		{
			sscanf(tokenptr, "%d", &start_channel);
			end_channel = start_channel;
		}
		else
		{
			sscanf(tokenptr, "%d:%d", &start_channel, &end_channel);
		}
					
		for(chan = start_channel; chan <= end_channel; chan++) 
		{
			checkErr(m220_RouteOpenQuery(handle, chan, &opened));
			
			// if (opened), append 1-comma to list, otherwise append 0-comma
			if(opened)
			{
				sprintf(str, "%d,", 1);
				strcat(route_open_list, str);
			}
			else
			{
				sprintf(str, "%d,", 0);
				strcat(route_open_list, str);
			}
		}
						
		tokenptr = strtok(NULL, ",");
	}

Error:
	return error;
}




/***************************************************************************** 
*  m220_ChannelStatus
*
* Find status of channel bit in relay reg
******************************************************************************/
Int32 m220_ChannelStatus(M220_HANDLE handle, UInt16 channel, Bool *status)
{   
	Int32 error = M220_SUCCESS;
	UInt16 offset, bit_mask, reg; 
	
	checkParamPointer(handle, 1);
	checkParamRange(channel, 0, 15, 2);
	checkParamPointer(status, 3);
	
	bit_mask = (0x01 << channel%4);
	offset = M220_ROW0_SET_REGISTER + ((channel/4) * 4);          									
	

	checkErr(m220_ReadRegister (handle, offset, &reg));
	
	if(reg & bit_mask)
		*status = M220_TRUE;
	else
		*status = M220_FALSE;
	
Error:
	return error;
}


/***************************************************************************** 
*  m220_Overlap
*
******************************************************************************/
m220_Overlap(M220_HANDLE handle, Bool overlap)
{ 
	Int32 error = M220_SUCCESS;
	
	checkParamPointer(handle, 1);
	checkParamRange(overlap, 0, 1, 2);

	handle->overlap = overlap;
		
Error:
		return error;
}



/***************************************************************************** 
*  m220_OverlapQuery
*
* Gets overlap mode. 
*                    If overlap = 1 functions exit immediately
*                       overlap = 0 functions wait till settling time
*                                   of relays (set in timer mode) has elapsed.
******************************************************************************/
m220_OverlapQuery(M220_HANDLE handle, Bool *overlap)
{ 
	Int32 error = M220_SUCCESS;
	
	checkParamPointer(handle, 1);
	checkParamPointer(overlap, 2);

	*overlap = handle->overlap;
		
Error:
		return error;
}

/**********************************************************************************
* m220_SetAllQuery
*
* Queries state of each channel (from registers)
* IN: M220_HANDLE handle
* OUT: Int16 *state
***********************************************************************************/
m220_SetAllQuery(M220_HANDLE handle, Int16 *state)
{ 
	Int32 error = M220_SUCCESS;
	Int32 row;
	Int16 reg;
	
	
	checkParamPointer(handle, 1);
	checkParamPointer(state, 2);
	
	*state = 0;
	
	for(row=0; row<4; row++)
	{	
		checkErr(m220_ReadRegister (handle, M220_ROW0_SET_REGISTER + (0x4*row), &reg));
		reg &= 0x000F;
		*state |= reg << (0x4*row);
	}
	
	
	Error:
			return error;
}

/**************************************************************************************
* m220_SetAll
*
* Sets all relays to their state (1=closed, 0=open)  passed as Int16 state
* IN: M220_HANDLE handle
*			Int16 state
*OUT: none
****************************************************************************************/
m220_SetAll(M220_HANDLE handle, Int16 state)
{ 
	Int32 error = M220_SUCCESS;
	Int32 row;
	Int16 reg;
	
	checkParamPointer(handle, 1);
	
	for(row=0; row<4; row++)
	{		
		reg = (state >> (0x4 * row)) & 0x000F;
		checkErr(m220_WriteRegister (handle, M220_ROW0_SET_REGISTER + (0x4*row), reg));
		checkErr(m220_WriteRegister (handle, M220_ROW0_RESET_REGISTER + (0x4*row), reg));
	}
	
Error:
	return error;
}


/*****************************************************************************
*	Modify the functions below this point to use the M220 driver in an 
*	environment not supported by an APIS platform support module. All functions
*	above this point should not have to be modified.
*
*	In general, any call to an APIS function (apis_func()) will need to be
*	modified to communicate with the instrument or operating system being used.
*
******************************************************************************/
/*****************************************************************************
*	m220_Init
*
******************************************************************************/
Int32 m220_Init (Int32 instr, Bool IDQuery, Bool reset, M220_HANDLE *handle)
{	
	Int32 error = M220_SUCCESS;
	Int32 warning = M220_SUCCESS;
  UInt16 value;

//  Check parameters  
	checkParamRange(IDQuery, 0, 1, 2);
	checkParamRange(reset, 0, 1, 3);
	checkParamPointer(handle, 4);

// Allocate the data structure
	(*handle) = 0;
	(*handle) = malloc(sizeof(M220_INSTR));
		if((*handle) == NULL)
	{	error = M220_ERROR_OUT_OF_MEMORY;
		goto Error;
	}
	
	(*handle)->apis_handle = 0;
	apisCheckErrInit(apis_open((APIS_PATH)instr, &(*handle)->apis_handle));

// Initialize Attributes 
  (*handle)->overlap = 1;
	(*handle)->fifo_full_wait=0;
	(*handle)->timeout_ms=20;       // 20 ms
  
// Start with power disabled to relays
	checkErr(m220_EnableRelayPower(*handle, 0));

// Perform ID Query 
	 if(IDQuery == M220_TRUE)
	{	checkErr(m220_GetIdentWord (*handle, 17, &value));
		value &= 0x0FFF;
		if (value != (UInt16)M220_VXI_MANF_ID) 
		{	error = M220_ERROR_FAIL_ID_QUERY;
 			goto Error;
	  }

		checkErr(m220_GetIdentWord (*handle, 18, &value));
		value &= 0x0FFF;
		if (value != (UInt16)M220_VXI_MODEL_CODE) 
		{	error = M220_ERROR_FAIL_ID_QUERY;
			goto Error;
		} 
	}

// Reset Instrument 
  if (reset == M220_TRUE) 
  {	
		checkErr(m220_Reset (*handle));
  }

Error:
		
	if((error == M220_SUCCESS) && (warning != M220_SUCCESS))		// if no errors then return the warnings		
		error = warning;
								
	return error; 
}

/*****************************************************************************
*	m220_Close
*
*	Close the session to the instrument.
******************************************************************************/
Int32 m220_Close (M220_HANDLE handle)
{	
	Int32 error = M220_SUCCESS;

	checkParamPointer(handle, 1);
	
	if(handle->apis_handle != NULL)
		apisCheckErr(apis_close(handle->apis_handle));
	free(handle);
Error:
	return error;
}

/*****************************************************************************
*	m220_ReadRegister
*
*	Read the specified offset from the module's I/O space and returns
*	the result.
******************************************************************************/
Int32 m220_ReadRegister (M220_HANDLE handle, UInt16 offset, UInt16 *value)
{	
	Int32 error = M220_SUCCESS;

	checkParamPointer(handle, 1);
	checkParamPointer(value, 3);

	apisCheckErr(apis_read(handle->apis_handle, 2, offset, value));
	
Error:
	return error;
}

/***************************************************************************
*	m220_WriteRegister
*
*	Writes to the specified offset in the module's I/O space
****************************************************************************/
Int32 m220_WriteRegister(M220_HANDLE handle, UInt16 offset, UInt16 value)
{	
	Int32 error = M220_SUCCESS;
	Bool op_complete = 0;
	Bool fifo_full = 0;
	UInt16 t_elapsed_ms = 0; 
	APIS_TICK t0;
	
	checkParamPointer(handle, 1);
	
	if(handle->fifo_full_wait)
	{	
		checkErr(m220_FIFOFullStatus(handle, &fifo_full));
		t0 = apis_tick();
		t_elapsed_ms = 0;
		while(fifo_full  && (t_elapsed_ms <= handle->timeout_ms)) 
		{ checkErr(m220_FIFOFullStatus(handle, &fifo_full));
			t_elapsed_ms = (apis_tick() - t0)*1000/APIS_TICKS_PER_SEC;
		}
		if(t_elapsed_ms >= handle->timeout_ms) 
		{	
			error = M220_ERROR_WAIT_TIMEOUT_EXPIRED;
			goto Error;
		}
	}
	
	if(handle->overlap == 0)
	{
		t0 = apis_tick();
		t_elapsed_ms = 0;
		while( (!op_complete) && (t_elapsed_ms <= handle->timeout_ms) ) 
		{ 
			checkErr(m220_OpcQuery(handle, &op_complete));  // If overlap = 0, do not return until FIFO is empty , or
			t_elapsed_ms = (apis_tick() - t0)*1000/APIS_TICKS_PER_SEC;
		}
				
		if(t_elapsed_ms >= handle->timeout_ms) 
		{	
			error = M220_ERROR_WAIT_TIMEOUT_EXPIRED;
			goto Error;
		}
	}
	apisCheckErr(apis_write(handle->apis_handle, 2, offset, value));

Error:
	return error;
}

/*****************************************************************************
*	m220_Reset 
*
******************************************************************************/
Int32 m220_Reset(M220_HANDLE handle)
{	
	Int32 error = M220_SUCCESS;
	UInt16 control_reg;
	UInt16 data;
	
	checkParamPointer(handle, 1);

	checkErr(m220_ReadRegister (handle, M220_CONTROL_REGISTER, &control_reg));
	control_reg |= M220_CONTROL_RESET_BIT;                                 				
	checkErr(m220_WriteRegister (handle, M220_CONTROL_REGISTER, control_reg));
	control_reg &= ~M220_CONTROL_RESET_BIT;
	checkErr(m220_WriteRegister (handle, M220_CONTROL_REGISTER, control_reg));
	
Error:
	return error;
}

/*****************************************************************************
*	m220_InstallIrq
*	
*	Installs the interrupts handler to be called when an interrupt occurs
*	and configures VISA to handle the interrutps by calling the handler.
******************************************************************************/
Int32 m220_InstallIrq(M220_HANDLE handle, void *interruptHandler, Int32 vector, Int32 level, Int32 mode, void *var_ptr)
{	
	Int32 error = M220_SUCCESS;

/*  Check parameters  */
	checkParamPointer(handle, 1);
	checkParamPointer(interruptHandler, 2);

apisCheckErr(apis_irqinstall(handle->apis_handle, interruptHandler, vector, level, mode, var_ptr));

Error:
	return error;
}


/*****************************************************************************
*	m220_UninstallIrq
*
*	Un-installs the interrupt handler for the instrument and disables
*	VISA interrupt handeling for the instrument
******************************************************************************/
Int32 m220_UninstallIrq(M220_HANDLE handle)
{	
	Int32 error = M220_SUCCESS;

	checkParamPointer(handle, 1);
	
	apisCheckErr(apis_irqremove(handle->apis_handle));

Error:
	return error;
}

/*****************************************************************************
*	m220_GetApisError
*
*	Returns the saved apis error code
******************************************************************************/
Int32 m220_GetApisError (M220_HANDLE handle, UInt32 *apis_error)
{	
	Int32 error = M220_SUCCESS;

	checkParamPointer(handle, 1);
	checkParamPointer(apis_error, 2);

	*apis_error = handle->apis_error;
	
Error:	
	return error;
}

